import os
import glob
import json
import time
import shutil

from . import _utils

IS_FILE_INPUT = _utils.META_INDEX["IS_FILE_INPUT"]


def start_loop():
    _utils.META_INDEX["last_prediction_loop_start_time"] = 0

    ACCEPTS_EXTRAS = False
    try:
        from predictor import predictor

        try:
            predictor(_utils.example[:1], extras=[None])
            ACCEPTS_EXTRAS = True
        except:
            pass

        _utils.META_INDEX["ACCEPTS_EXTRAS"] = ACCEPTS_EXTRAS

        _utils.logger.info(f'ACCEPTS_EXTRAS: {_utils.META_INDEX["ACCEPTS_EXTRAS"]}')
    except Exception as ex:
        _utils.logger.exception(ex, exc_info=True)

    """
    The Prediction loop. This is where the logic happens.

    This function starts a loop. Does not return anything.
    """

    # warmup
    _utils.warmup(predictor, _utils.example)

    # find optimal batch size and get_time_per _utils.example
    batch_size, time_per_example = _utils.find_optimum_batch_sizes(
        predictor, _utils.example
    )

    max_wait_time = time_per_example * _utils.MAX_WAIT

    # write batch size to temp file for use in generating _run.sh
    _utils.META_INDEX["batch_size"] = batch_size
    _utils.META_INDEX["time_per_example"] = time_per_example
    _utils.logger.info(f"max_wait_time: {max_wait_time}, batch_size: {batch_size}")

    # list of files/data to be processed is tracked here.
    to_process = None

    _utils.logger.info("Starting prediction loop")

    while True:
        # Get the latest list of to process data
        batch = []

        batch_extra_options = []

        unique_ids = []
        unique_id_to_metrics = {}
        batch_collection_start_time = time.time()
        first_sleep_start_time = 0
        __loop_is_sleeping = False
        while True:
            _utils.META_INDEX["last_prediction_loop_start_time"] = time.time()
            
            if len(_utils.REQUEST_INDEX):
                (
                    unique_id,
                    (in_data, unique_id_to_metrics[unique_id], _batch_extra_options),
                ) = _utils.REQUEST_INDEX.popitem(last=False)
                batch_collection_start_time = time.time()

                for __i, _ in enumerate(in_data):
                    unique_ids.append(unique_id)
                    batch.append(_)

                batch_extra_options += _batch_extra_options

            if len(unique_ids) == 0:
                if first_sleep_start_time == 0:
                    first_sleep_start_time = time.time()
                else:
                    if (
                        time.time() - first_sleep_start_time
                        >= _utils.BATCH_COLLECTION_SLEEP_IF_EMPTY_FOR
                    ):
                        if not __loop_is_sleeping:
                            _utils.logger.info(
                                f"Empty for {_utils.BATCH_COLLECTION_SLEEP_IF_EMPTY_FOR} sec, loop sleep started. wakeup_interval: {_utils.BATCH_COLLECTION_SLEEP_FOR_IF_EMPTY}"
                            )
                        __loop_is_sleeping = True
                        time.sleep(_utils.BATCH_COLLECTION_SLEEP_FOR_IF_EMPTY)
                        continue

                time.sleep(_utils.PREDICTION_LOOP_SLEEP)
                continue

            first_sleep_start_time = 0
            __loop_is_sleeping = False

            if len(batch) >= batch_size:
                unique_id_count = len(set(unique_ids))
                if unique_id_count > 1:
                    _utils.logger.info(
                        f"Batch of size: {len(batch)}, max_batch_size: {batch_size}, unique_ids: {unique_id_count} collected."
                    )
                batch_collection_start_time = 0
                break

            if batch_collection_start_time != 0 and (
                time.time() - batch_collection_start_time >= max_wait_time
            ):
                unique_id_count = len(set(unique_ids))
                if unique_id_count > 1:
                    _utils.logger.info(
                        f"Batch of size: {len(batch)}, max_batch_size: {batch_size}, unique_ids: {unique_id_count} collected."
                    )
                batch_collection_start_time = 0
                break

            time.sleep(time_per_example * 0.004)

        try:
            pred_start_time = time.time()
            __in_batch_length = len(batch)

            if ACCEPTS_EXTRAS:
                preds = predictor(
                    batch, batch_size=batch_size, extras=batch_extra_options
                )
            else:
                preds = predictor(batch, batch_size=batch_size)

            if not isinstance(preds, list) or len(preds) != __in_batch_length:
                _utils.logger.error(
                    "Something is seriously wrong! len(inputs) != len(outputs) from predictor.py. Check your recipe"
                )
                _utils.logger.error(
                    f"Inputs: {batch} length: {__in_batch_length, len(batch)}"
                )
                _utils.logger.error(
                    f"Preds: {preds} length: {len(preds) if isinstance(preds, list) else 'N/A'}"
                )
                exit()

            pred_end_time = time.time()

            for unique_id in unique_id_to_metrics:
                unique_id_to_metrics[unique_id]["prediction_start"] = pred_start_time
                unique_id_to_metrics[unique_id]["prediction_end"] = pred_end_time
                unique_id_to_metrics[unique_id]["predicted_in_batch"] = len(unique_ids)

            _utils.logger.info(
                f"Batch of size: {len(batch)}, max_batch_size: {batch_size} predicted."
            )
        except Exception as ex:
            _utils.logger.exception(ex, exc_info=True)
            preds = [{"success": False, "reason": str(ex)} for _ in range(len(batch))]

        if IS_FILE_INPUT:
            for _ in batch:
                try:
                    os.remove(_)
                except Exception as ex:
                    _utils.logger.warning(ex, exc_info=True)

        unique_id_wise_results = {}
        for unique_id, pred in zip(unique_ids, preds):
            if unique_id not in unique_id_wise_results:
                unique_id_wise_results[unique_id] = []

            unique_id_wise_results[unique_id].append(pred)

        for unique_id, _preds in unique_id_wise_results.items():
            _utils.RESULTS_INDEX[unique_id] = (_preds, unique_id_to_metrics[unique_id])

        time.sleep(_utils.PREDICTION_LOOP_SLEEP)


if __name__ == "__main__":
    start_loop()



import resource

try:
    resource.setrlimit(resource.RLIMIT_NOFILE, (131072, 131072))
except:
    pass

import os
import sys
import subprocess

RECIPE = os.getenv("RECIPE")
MODE = os.getenv("MODE")
QUEUE_DIR = os.getenv("QUEUE_DIR")
QUEUE_NAME = os.getenv(f"QUEUE_NAME", f"default")
WORKERS = int(os.getenv("WORKERS", 3))
TIMEOUT = os.getenv("TIMEOUT", "1000")

if not RECIPE or not MODE:
    import argparse

    parser = argparse.ArgumentParser(description="CLI for fastDeploy")
    parser.add_argument(
        "--recipe",
        type=str,
        help="Path to your recipe folder; env: RECIPE",
        required=True,
    )

    parser.add_argument(
        "--mode",
        type=str,
        help='One of, ["loop", "rest", "websocket", "build_rest"]; env: MODE',
        required=True,
    )

    parser.add_argument(
        "--queue_dir",
        type=str,
        help='Defaults to "--recipe" folder; end: QUEUE_DIR',
        required=False,
    )

    parser.add_argument(
        "--base",
        type=str,
        help='docker image to use as base. defaults to "auto"',
        required=False,
    )

    parser.add_argument(
        "--docker_args",
        type=str,
        help="This string will be passed to docker as args",
        required=False,
    )

    args = parser.parse_args()

    QUEUE_DIR = os.path.join(os.path.abspath(args.recipe), "fastdeploy_dbs")

    MODE = args.mode
    RECIPE = os.path.abspath(args.recipe)
    BASE = args.base
    DOCKER_ARGS = args.docker_args
    if not DOCKER_ARGS:
        DOCKER_ARGS = ""

if os.path.exists(RECIPE):
    sys.path.append(RECIPE)
    os.chdir(RECIPE)

if not QUEUE_DIR:
    QUEUE_DIR = os.path.join(RECIPE, "fastdeploy_dbs")

    try:
        if not os.path.exists(os.path.join(RECIPE, ".gitignore")):
            _gitignore_f = open(os.path.join(RECIPE, ".gitignore"), "a")
            _gitignore_f.write("\nfastdeploy_dbs\nfastdeploy_dbs/*\n")
            _gitignore_f.flush()
            _gitignore_f.close()
    except:
        pass


def loop():
    from ._loop import start_loop

    start_loop()


def rest():
    from ._app import app
    import gunicorn.app.base

    class StandaloneApplication(gunicorn.app.base.BaseApplication):
        def __init__(self, app, options=None):
            self.options = options or {}
            self.application = app
            super().__init__()

        def load_config(self):
            config = {
                key: value
                for key, value in self.options.items()
                if key in self.cfg.settings and value is not None
            }
            for key, value in config.items():
                self.cfg.set(key.lower(), value)

        def load(self):
            return self.application

    port = int(os.getenv("PORT", "8080"))
    host = os.getenv("HOST", "0.0.0.0")

    options = {
        "preload": "",
        "bind": "%s:%s" % (host, port),
        "workers": WORKERS,
        "worker_connections": 1000,
        "worker_class": "gevent",
        "timeout": TIMEOUT,
    }

    print(f"fastDeploy active at http://{host}:{port}")

    StandaloneApplication(app, options).run()


def websocket():
    from ._app import WebSocketInfer
    from collections import OrderedDict
    from geventwebsocket import WebSocketServer, Resource

    port = int(os.getenv("PORT", "8080"))
    host = os.getenv("HOST", "0.0.0.0")

    WebSocketServer(
        (host, port), Resource(OrderedDict([("/infer", WebSocketInfer)]))
    ).serve_forever()


def build(mode="build_rest"):
    dockerfile_lines = []

    if not BASE:
        base = "python:3.6-slim"
    else:
        base = BASE

    dockerfile_lines.append(f"FROM {base}")
    dockerfile_lines.append(
        f"RUN python3 -m pip install --upgrade --no-cache-dir pip fastdeploy"
    )

    # if not recipe_base_name:
    recipe_base_name = "/recipe"

    dockerfile_lines.append(f"ADD . {recipe_base_name}")

    if os.path.exists(os.path.join(RECIPE, "extras.sh")):
        dockerfile_lines.append(f"RUN cd {recipe_base_name} && bash extras.sh")

    dockerfile_lines.append(
        f"RUN cd {recipe_base_name} && python3 -m pip install --no-cache-dir -r requirements.txt"
    )

    

    dockerfile_lines.append(
        f"RUN sudo chmod -R a+rw {recipe_base_name} || chmod -R a+rw {recipe_base_name} && cd {recipe_base_name} && python3 predictor.py"
    )

    dockerfile_lines.append(
        f'ENTRYPOINT {os.getenv("ENTRYPOINT", ["/bin/sh", "-c"])} \n'.replace("'", '"')
    )

    if mode == "build_rest":
        dockerfile_lines.append(
            f'CMD ["ulimit -n 1000000 && python3 -m fastdeploy --recipe {recipe_base_name} --mode loop & python3 -m fastdeploy --recipe {recipe_base_name} --mode rest"] \n'
        )
    elif mode == "build_websocket":
        dockerfile_lines.append(
            f'CMD ["ulimit -n 1000000 && python3 -m fastdeploy --recipe {recipe_base_name} --mode loop & python3 -m fastdeploy --recipe {recipe_base_name} --mode websocket"] \n'
        )

    dockerfile_path = os.path.join(RECIPE, "fastDeploy.auto_dockerfile")
    _dockerignore_f = open(os.path.join(RECIPE, ".dockerignore"), "w")
    _dockerignore_f.write("*.request_index\n*.results_index\n*.log_index")
    _dockerignore_f.flush()
    _dockerignore_f.close()

    _gitignore_f = open(os.path.join(RECIPE, ".gitignore"), "a")
    _gitignore_f.write("\nfastdeploy_dbs\nfastdeploy_dbs/*\n")
    _gitignore_f.flush()
    _gitignore_f.close()

    _f = open(dockerfile_path, "w")
    _f.write("\n".join(dockerfile_lines))
    _f.flush()
    _f.close()

    docker_image_name = f"fastdeploy_{os.path.basename(os.path.abspath(RECIPE))}".strip(
        "/"
    ).lower()

    subprocess.call(
        f"docker build {DOCKER_ARGS} -f {dockerfile_path} -t {docker_image_name} {RECIPE}",
        shell=True,
    )

    print(f"{docker_image_name} built!")
    exit()


if MODE == "loop":
    loop()

elif MODE == "rest":

    rest()

elif MODE == "build_rest":
    build(mode="build_rest")

elif MODE == "build_websocket":
    build(mode="build_socket")

elif MODE == "websocket":
    websocket()



import logging

logging.basicConfig(
    format="%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s",
    datefmt="%Y-%m-%d:%H:%M:%S",
    level=logging.INFO,
)

logger = logging.getLogger(__name__)


import os
import uuid
import glob
import json
import time
import shlex
import shutil
from datetime import datetime
from diskcache import Index, Cache

import sys
from example import example

IS_FILE_INPUT = False
try:
    IS_FILE_INPUT = os.path.exists(example[0])
except:
    pass

from . import QUEUE_DIR, QUEUE_NAME

# En variable to configure allowed origins
ALLOWED_ORIGINS = os.getenv("ALLOWED_ORIGINS", "*")

PREDICTION_LOOP_SLEEP = float(os.getenv("PREDICTION_LOOP_SLEEP", "0.06"))
BATCH_COLLECTION_SLEEP_IF_EMPTY_FOR = float(
    os.getenv("BATCH_COLLECTION_SLEEP_IF_EMPTY_FOR", "60")
)
BATCH_COLLECTION_SLEEP_FOR_IF_EMPTY = float(
    os.getenv("BATCH_COLLECTION_SLEEP_FOR_IF_EMPTY", "1")
)
MANAGER_LOOP_SLEEP = float(os.getenv("MANAGER_LOOP_SLEEP", "8"))

_request_index = os.path.join(QUEUE_DIR, f"{QUEUE_NAME}.request_index")
_results_cache = os.path.join(QUEUE_DIR, f"{QUEUE_NAME}.results_cache")
_metrics_cache = os.path.join(QUEUE_DIR, f"{QUEUE_NAME}.metrics_cache")
_meta_index = os.path.join(QUEUE_DIR, f"{QUEUE_NAME}.META_INDEX")

REQUEST_INDEX = Index(_request_index)
RESULTS_INDEX = Cache(_results_cache)
METRICS_CACHE = Cache(_metrics_cache)
META_INDEX = Index(_meta_index)

META_INDEX["IS_FILE_INPUT"] = IS_FILE_INPUT
META_INDEX["PREDICTION_LOOP_SLEEP"] = PREDICTION_LOOP_SLEEP
META_INDEX["BATCH_COLLECTION_SLEEP_IF_EMPTY_FOR"] = BATCH_COLLECTION_SLEEP_IF_EMPTY_FOR
META_INDEX["BATCH_COLLECTION_SLEEP_FOR_IF_EMPTY"] = BATCH_COLLECTION_SLEEP_FOR_IF_EMPTY
META_INDEX["MANAGER_LOOP_SLEEP"] = MANAGER_LOOP_SLEEP
META_INDEX["TOTAL_REQUESTS"] = 0

FASTDEPLOY_UI_PATH = os.getenv(
    "FASTDEPLOYUI",
    os.path.join(os.path.split(os.path.abspath(__file__))[0], "fastdeploy-ui"),
)

logger.info(
    f"REQUEST_INDEX: {_request_index} RESULTS_INDEX: {_results_cache} META_INDEX: {_meta_index} IS_FILE_INPUT: {IS_FILE_INPUT} FASTDEPLOY_UI_PATH: {FASTDEPLOY_UI_PATH}"
)

# clear if not
# No real use in making these configurable.

# Number of gunicorn workers to use
# Keep 0 for auto selection
WORKERS = int(os.getenv("WORKERS", "0"))

TIMEOUT = int(os.getenv("TIMEOUT", "120"))

# if BATCH_SIZE is not 0, will be used as default batch size.
BATCH_SIZE = int(os.getenv("BATCH_SIZE", "0"))

# Maximum examples allowed in client batch.
# 0 means unlimited
MAX_PER_CLIENT_BATCH = int(os.getenv("MAX_PER_CLIENT_BATCH", "0"))

# The loop will wait for time_per_example * MAX_WAIT for batching.
MAX_WAIT = float(os.getenv("MAX_WAIT", 0.2))


def warmup(predictor, example_input, n=3):
    """
    Run warmup prediction on the model.

    :param n: number of warmup predictions to be run. defaults to 3
    """
    logger.info("Warming up .. ")
    for _ in range(n):
        predictor(example_input)


def find_optimum_batch_sizes(
    predictor,
    example_input,
    max_batch_search_sec=int(os.getenv("MAX_BATCH_SEARCH_SEC", "240")),
):
    """
    Finds the optimum batch size for a predictor function with the given example input.

    :param predictor: predictor function. Should have two inputs, a list of examples and batch size.
    :param example_input: example input for the predictor.
    :param max_batch_search_sec: max time to spend on batch size search in seconds.

    :return batch_size: optimal batch size to be used
    :return time_per_example: approx time taken per example.
    """
    time_per_example = None
    previous_time_per_example = pow(2, 64)

    possible_batch_sizes = range(16)
    if BATCH_SIZE:
        possible_batch_sizes = [BATCH_SIZE]

    if len(possible_batch_sizes) > 1:
        possible_batch_sizes = [
            pow(2, batch_size) for batch_size in possible_batch_sizes
        ]

    search_start_time = time.time()
    batch_size_to_time_per_example = {}
    for b_i, batch_size in enumerate(possible_batch_sizes):
        start = time.time()
        if start - search_start_time >= max_batch_search_sec:
            batch_size = possible_batch_sizes[b_i - 1]
            logger.warn(
                f"Batch size set to {batch_size} because of MAX_BATCH_SEARCH_SEC: {max_batch_search_sec}"
            )
            break
        try:
            for _ in range(3):
                inputs = example_input * batch_size
                inputs = inputs[:batch_size]

                preds = predictor(inputs, batch_size=batch_size)
                if not isinstance(preds, list) or len(preds) != batch_size:
                    logger.error(
                        "Something is seriously wrong! len(inputs) != len(outputs) from predictor.py. Check your recipe"
                    )
                    logger.error(f"Inputs: {inputs} length: {batch_size}")
                    logger.error(
                        f"Preds: {preds} length: {len(preds) if isinstance(preds, list) else 'N/A'}"
                    )
                    exit()
        except Exception as ex:
            logger.exception(ex, exc_info=True)
            logger.warn("Batch size set to 1 because of above exception")
            break

        time_per_example = (time.time() - start) / (3 * batch_size)

        logger.info(
            f"Time per sample for batch_size: {batch_size} is {time_per_example}"
        )

        batch_size_to_time_per_example[batch_size] = time_per_example

        META_INDEX[f"batch_size_to_time_per_example"] = batch_size_to_time_per_example

        # determine which batch size yields the least time per example.

        if time_per_example > previous_time_per_example * 0.95:
            break
        else:
            previous_time_per_example = time_per_example

    if not BATCH_SIZE:
        batch_size = int(max(1, batch_size / 2))
    else:
        batch_size = BATCH_SIZE

    logger.info(f"optimum batch size is {batch_size}")

    return batch_size, time_per_example


def write_webhook(unique_id, webhook):
    """
    writes webhook string (url) to corresponding file.

    :param unique_id: unique_id
    :param webhook: webhook string
    """
    if webhook and isinstance(webhook, str):
        open(os.path.join(RAM_DIR, unique_id + ".webhook"), "w").write(webhook)
    else:
        if webhook is not None:
            logger.warn(f"id: {unique_id}, webhook: {webhook} is not valid.")



from gevent import monkey

monkey.patch_all()
import gevent
import gevent.pool
import os
import sys
import glob
import time
import uuid
import ujson
import falcon
import base64
import shutil
import logging
import datetime
import mimetypes
from functools import partial

from . import _utils

_utils.logger.info(f"Waiting for warmup, batch size search to finish.")
while "time_per_example" not in _utils.META_INDEX:
    time.sleep(5)

ONLY_ASYNC = os.getenv("ONLY_ASYNC", False)

TIME_PER_EXAMPLE = _utils.META_INDEX["time_per_example"]
IS_FILE_INPUT = _utils.META_INDEX["IS_FILE_INPUT"]


def wait_and_read_pred(unique_id):
    """
    Waits for and reads result for unique_id.

    :param unique_id: unique_id of the input

    :return response: json dumped python dict with keys "success" and "prediction"/ "reason"
    :return status: HTTP status code
    """
    # Keeping track of start_time for TIMEOUT implementation
    start_time = time.time()
    # Default response and status
    response, status = (
        {"success": False, "reason": "timeout"},
        falcon.HTTP_503,
    )
    while True:
        try:
            # if result doesn't exist for this uuid,  while loop continues/
            pred, metrics = _utils.RESULTS_INDEX[unique_id]
            try:
                response = {"prediction": pred, "success": True}
            # if return dict has any non json serializable values, we str() it
            except:
                _utils.logger.info(
                    f"unique_id: {unique_id} could not json serialize the result."
                )
                response = {"prediction": str(pred), "success": True}
            status = falcon.HTTP_200
            break
        except:
            # stop in case of timeout
            if time.time() - start_time >= _utils.TIMEOUT:
                _utils.logger.warn(
                    f"unique_id: {unique_id} timedout, with timeout {_utils.TIMEOUT}"
                )
                break

            gevent.time.sleep(TIME_PER_EXAMPLE * 0.501)
            metrics = {}

    return response, status, metrics


class Infer(object):
    def on_post(self, req, resp):
        try:
            unique_id = str(uuid.uuid4())

            req_params = req.params
            is_async_request = ONLY_ASYNC or req_params.get("async")

            _extra_options_for_predictor = {}

            if (req.content_type == "application/json" and IS_FILE_INPUT) or (
                req.content_type != "application/json" and not IS_FILE_INPUT
            ):
                if IS_FILE_INPUT:
                    resp.media = {
                        "success": False,
                        "reason": f"Received json input. Expected multi-part file input.",
                    }
                else:
                    resp.media = {
                        "success": False,
                        "reason": f"Received multi-part file input. Expected json input.",
                    }

                resp.status = falcon.HTTP_400

            else:
                if req.content_type == "application/json":
                    in_data = req.media

                    try:
                        # Legacy. use data in "data" key if exists
                        in_data = in_data["data"]
                    except:
                        pass

                    _in_file_names = [None for _ in range(len(in_data))]

                else:
                    in_data = []
                    _in_file_names = []

                    for part in req.get_media():
                        if not part.filename and _utils.META_INDEX["ACCEPTS_EXTRAS"]:
                            try:
                                _extra_options_for_predictor.update(
                                    ujson.loads(part.text)
                                )
                            except:
                                pass

                        else:
                            _in_file_names.append(part.name)

                            _temp_file_path = (
                                f"{uuid.uuid4()}{os.path.splitext(part.filename)[1]}"
                            )
                            _temp_file = open(_temp_file_path, "wb")

                            while True:
                                chunk = part.stream.read(2048)
                                if not chunk:
                                    break

                                _temp_file.write(chunk)
                            _temp_file.flush()
                            _temp_file.close()

                            in_data.append(_temp_file_path)

                metrics = {
                    "received": time.time(),
                    "prediction_start": -1,
                    "prediction_end": -1,
                    "batch_size": len(in_data),
                    "predicted_in_batch": -1,
                    "responded": -1,
                }

                _utils.REQUEST_INDEX[unique_id] = (
                    in_data,
                    metrics,
                    [_extra_options_for_predictor.get(_) for _ in _in_file_names],
                )

                _utils.META_INDEX["TOTAL_REQUESTS"] += 1

                if is_async_request:
                    resp.media = {"unique_id": unique_id, "success": True}
                    resp.status = falcon.HTTP_200
                else:
                    preds, status, _metrics = wait_and_read_pred(unique_id)

                    if not len(_metrics):
                        _metrics = metrics

                    _metrics["responded"] = time.time()
                    _utils.METRICS_CACHE[len(_utils.METRICS_CACHE)] = (
                        unique_id,
                        _metrics,
                        in_data,
                    )

                    resp.media = preds
                    resp.status = status

        except Exception as ex:
            _utils.logger.exception(ex, exc_info=True)
            resp.media = {"success": False, "reason": "malformed request"}
            resp.status = falcon.HTTP_400


class Metrics(object):
    def on_get(self, req, resp):

        try:
            end_time = int(req.params.get("from_time", time.time()))
            total_time = int(req.params.get("total_time", 3600))

            loop_batch_size = _utils.META_INDEX["batch_size"]
            batch_size_to_time_per_example = _utils.META_INDEX[
                "batch_size_to_time_per_example"
            ]

            first_end_time = 0
            all_metrics_in_time_period = {
                "time_graph_data": {
                    "labels": [],
                    "datasets": [
                        {"name": "Response time", "values": []},
                        {"name": "Prediction time", "values": []},
                    ],
                },
                "auto_batching_graph_data": {
                    "labels": [],
                    "datasets": [
                        {"name": "Input batch size", "values": []},
                        {"name": "Dynamically batched to", "values": []},
                    ],
                },
                "index_to_all_meta": {},
            }

            current_time = time.time()

            n_metrics = len(_utils.METRICS_CACHE)
            total_requests = n_metrics

            for _ in reversed(range(n_metrics)):
                unique_id, _metrics, _in_data = _utils.METRICS_CACHE[_]
                # max 5 second loop alowed
                if time.time() - current_time >= 5:
                    break

                received_time = _metrics["received"]
                prediction_start = _metrics["prediction_start"]
                prediction_end = _metrics["prediction_end"]
                batch_size = _metrics["batch_size"]
                predicted_in_batch = _metrics["predicted_in_batch"]
                responded_at = _metrics["responded"]

                prediction_time_per_example = (
                    prediction_end - prediction_start
                ) / predicted_in_batch

                if current_time - received_time >= total_time:
                    break

                if received_time <= 0 or responded_at <= 0:
                    continue

                x_id = total_requests - len(
                    all_metrics_in_time_period["index_to_all_meta"]
                )

                all_metrics_in_time_period["auto_batching_graph_data"]["labels"].insert(
                    0, x_id
                )
                all_metrics_in_time_period["auto_batching_graph_data"]["datasets"][0][
                    "values"
                ].insert(0, batch_size)
                all_metrics_in_time_period["auto_batching_graph_data"]["datasets"][1][
                    "values"
                ].insert(0, predicted_in_batch)

                all_metrics_in_time_period["time_graph_data"]["labels"].insert(0, x_id)
                all_metrics_in_time_period["time_graph_data"]["datasets"][0][
                    "values"
                ].insert(0, responded_at - received_time)
                all_metrics_in_time_period["time_graph_data"]["datasets"][1][
                    "values"
                ].insert(
                    0,
                    batch_size
                    * (prediction_end - prediction_start)
                    / predicted_in_batch,
                )

                all_metrics_in_time_period["index_to_all_meta"][
                    n_metrics - len(all_metrics_in_time_period["index_to_all_meta"])
                ] = {
                    "unique_id": unique_id,
                    "received_time": str(
                        datetime.datetime.fromtimestamp(received_time)
                    ),
                    "prediction_time_per_example": prediction_time_per_example,
                    "batch_size": batch_size,
                    "start_to_end_time": responded_at - received_time,
                    "predicted_in_batch": predicted_in_batch,
                }

            resp.media = all_metrics_in_time_period
            resp.status = falcon.HTTP_200

        except Exception as ex:
            logging.exception(ex, exc_info=True)
            pass


ALL_META = {}
for k, v in _utils.META_INDEX.items():
    ALL_META[k] = v

ALL_META["is_file_input"] = IS_FILE_INPUT
ALL_META["example"] = _utils.example


class Health(object):
    def on_get(self, req, resp):
        stuck_for = float(req.params.get("stuck"))
        last_prediction_loop_start_time = _utils.META_INDEX["last_prediction_loop_start_time"]
        prediction_loop_stuck_for = time.time() - last_prediction_loop_start_time

        if last_prediction_loop_start_time:
            if (
                stuck_for
                and prediction_loop_stuck_for >= stuck_for
            ):
                resp.status = falcon.HTTP_503
                resp.media = {
                    "predictor_status": f"prediction loop stuck for {prediction_loop_stuck_for}. deemed stuck."
                }
            else:
                resp.status = falcon.HTTP_200
                resp.media = {
                    "predictor_status": f"prediction loop running for {prediction_loop_stuck_for}"
                }
        else:
            resp.status = falcon.HTTP_200
            resp.media = {
                "predictor_status": f"prediction loop not started"
            }


class Meta(object):
    def on_get(self, req, resp):
        if req.params.get("example") == "true":
            resp.content_type = mimetypes.guess_type(_utils.example[0])[0]
            resp.stream = open(_utils.example[0], "rb")
            resp.downloadable_as = os.path.basename(_utils.example[0])

        else:
            resp.media = ALL_META
            resp.status = falcon.HTTP_200


class Res(object):
    def on_post(self, req, resp):
        try:
            unique_id = req.media["unique_id"]
            _utils.logger.info(f"unique_id: {unique_id} Result request received.")
            try:
                pred, metrics = _utils.RESULTS_INDEX[unique_id]
                resp.media = {"success": True, "prediction": pred}
            except:
                if unique_id in _utils.REQUEST_INDEX:
                    resp.media = {"success": None, "reason": "processing"}
                else:
                    resp.media = {
                        "success": False,
                        "reason": "No request found with this unique_id",
                    }

            resp.status = falcon.HTTP_200

        except Exception as ex:
            _utils.logger.exception(ex, exc_info=True)
            resp.media = {"success": False, "reason": str(ex)}
            resp.status = falcon.HTTP_400


json_handler = falcon.media.JSONHandler(
    loads=ujson.loads, dumps=partial(ujson.dumps, ensure_ascii=False)
)

extra_handlers = {
    "application/json": json_handler,
}

app = falcon.App(cors_enable=True)
app.req_options.media_handlers.update(extra_handlers)
app.resp_options.media_handlers.update(extra_handlers)
app.req_options.auto_parse_form_urlencoded = True
app = falcon.App(
    middleware=falcon.CORSMiddleware(
        allow_origins=_utils.ALLOWED_ORIGINS, allow_credentials=_utils.ALLOWED_ORIGINS
    )
)

infer_api = Infer()
res_api = Res()
metrics_api = Metrics()
meta_api = Meta()
health_api = Health()

app.add_route("/infer", infer_api)
app.add_route("/result", res_api)
app.add_route("/metrics", metrics_api)
app.add_route("/meta", meta_api)
app.add_route("/health", health_api)


app.add_static_route(
    "/",
    _utils.FASTDEPLOY_UI_PATH,
    fallback_filename="index.html",
)

# Backwards compatibility
app.add_route("/sync", infer_api)

from geventwebsocket import WebSocketApplication


class WebSocketInfer(WebSocketApplication):
    def on_open(self):
        self.connection_id = f"{uuid.uuid4()}"
        self.n = 0
        self.start_time = time.time()
        _utils.logger.info(f"{self.connection_id} websocket connection opened.")

    def on_message(self, message):
        self.n += 1
        try:
            if message is not None:
                metrics = {
                    "received": time.time(),
                    "prediction_start": -1,
                    "prediction_end": -1,
                    "batch_size": 1,
                    "predicted_in_batch": -1,
                    "responded": -1,
                }
                message_id = f"{self.connection_id}.{self.n}"
                _utils.REQUEST_INDEX[message_id] = ([message], metrics)
                preds, status, metrics = wait_and_read_pred(message_id)
                if "prediction" in preds:
                    preds["prediction"] = preds["prediction"][0]

                self.ws.send(ujson.dumps(preds))

        except Exception as ex:
            _utils.logger.exception(ex, exc_info=True)
            pass

    def on_close(self, reason):
        _utils.logger.info(
            f"{self.connection_id} websocket connection closed. Time spent: {time.time() - self.start_time} n_mesages: {self.n}"
        )
        pass
